import requests
import pushy.util.localStorage as localStorage
from pushy.util.config import config

def getApiHost():
    # Development API endpoint for localhost
    # return config['api']['devEndpoint']
    
    # Attempt to fetch Pushy Enterprise hostname (may be null)
    enterpriseEndpoint = localStorage.get(config['storageKeys']['enterpriseEndpoint'])

    # Pushy Enterprise endpoint or fallback to Pushy Pro API endpoint
    return enterpriseEndpoint or config['api']['endpoint']

def post(path, json):
    # Build full URL to API endpoint
    url = getApiHost() + path

    # Execute HTTP request
    response = requests.post(url, json=json)

    # Invalid status code?
    if response.status_code < 200 or response.status_code > 299:
        # Convert response to JSON
        json = response.json()
        
        try:
            # Try extracting error code
            error = json['code'] + ': '
        except:
            error = ''

        try:
            # Try extracting error message
            error += json['error']
        except KeyError:
            error = 'An unknown error occurred'

        # Throw detailed error
        raise Exception(error + ' (HTTP status code ' + str(response.status_code) + ')')
    
    # Convert response to JSON and return it
    return response.json()


import json
import time
import ssl
import pushy
from pushy.util.config import config
import paho.mqtt.client as mqtt
import pushy.util.localStorage as localStorage

# Default client
client = mqtt.Client()

def getMqttEndpoint():
    # Attempt to fetch Pushy Enterprise hostname (fall back to Pushy Pro endpoint)
    endpoint = localStorage.get(config['storageKeys']['enterpriseMqttEndpoint']) or config['mqtt']['endpoint']

    # Get current unix timestamp (for timestamp injection)
    unix = str(int(time.time()))

    # Inject unix timestamp into hostname (if placeholder present)
    return endpoint.replace('{timestamp}', unix)

def getMqttPort():
    # Attempt to fetch Pushy Enterprise hostname to check if running in enterprise mode
    if localStorage.get(config['storageKeys']['enterpriseMqttEndpoint']):
        # Pushy Enterprise port number
        port = config['mqtt']['enterprisePortNumber']
    else:
        # Pushy Pro port number
        port = config['mqtt']['portNumber']

    # Return port number as an integer
    return int(port)

def getMqttKeepAlive():
    # Fetch custom keepalive interval (in seconds) or default to 300
    keepAliveInterval = localStorage.get(config['storageKeys']['keepAliveInterval']) or config['mqtt']['defaultKeepAlive']

    # Convert to integer
    return int(keepAliveInterval)

def setNotificationListener(listener):
    # Set global variable
    global notificationListener

    # Store reference to listener callback method
    notificationListener = listener

def setConnectionListener(listener):
    # Set global variable
    global connectionListener

    # Store reference to listener callback method
    connectionListener = listener

def setDisconnectionListener(listener):
    # Set global variable
    global disconnectionListener

    # Store reference to listener callback method
    disconnectionListener = listener

# Callback for when the MQTT client connects successfully
def on_connect(client, userdata, flags, rc):
    # Log connection success to console
    print('[Pushy] Connected successfully (device token ' + localStorage.get(config['storageKeys']['token']) + ')')

    # Check if connection listener defined & invoke it
    if connectionListener:
        connectionListener()

# Callback for when the MQTT client disconnects
def on_disconnect(client, userdata, msg):
    # Log disconnection to console
    print('[Pushy] Disconnected from server')
    
    # Check if disconnection listener defined & invoke it
    if disconnectionListener:
        disconnectionListener()

# Callback for when the MQTT client receives a notification
def on_message(client, userdata, msg):
    # Check if notification listener defined, otherwise do nothing
    if notificationListener:
        # Decode payload into UTF-8 string
        message = msg.payload.decode('utf-8')

        try:
            # Attempt to parse message into JSON
            payload = json.loads(message)

            # Invoke notification listener with JSON dictionary
            notificationListener(payload)
        except Exception as err:
            # Log error to console
            print('[Pushy] MQTT JSON Parse Error\n' + str(err))

def connect():
    # Device not registered yet?
    if not pushy.isRegistered():
        return

    # If client socket exists, don't recreate it
    # A disconnected client will reconnect automatically
    if client._sock:
        return
    
    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])

    # Reinitialize client with Pushy device token
    client.reinitialise(client_id=token)

    # Reset event listeners after reinitializing
    client.on_connect = on_connect
    client.on_message = on_message
    client.on_disconnect = on_disconnect

    # Set username and password accordingly
    client.username_pw_set(token, password=tokenAuth)

    # Enforce TLS encryption
    client.tls_set(ca_certs=None, certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLS, ciphers=None)
    
    try:
        # Try connecting to Pushy MQTT endpoint
        client.connect(getMqttEndpoint(), getMqttPort(), getMqttKeepAlive())
    except Exception as err:
        print('[Pushy] Connection failed with error: ' + str(err))
    
def disconnect():
    # Disconnect if client socket is connected
    if client._sock:
        client.disconnect()

def loop_forever():
    # Blocking call that processes network traffic, dispatches callbacks and
    # handles reconnecting
    # Other loop*() functions are available that give a threaded interface and a
    # manual interface
    if client._sock:
        client.loop_forever()


import dbm
from pathlib import Path

# Make sure 'db' subdirectory exists
Path('db').mkdir(exist_ok=True)

# Open and/or create key value store
db = dbm.open('db/pushy', 'c')

# Getter method
def get(key):
    try:
        return db[key].decode('utf-8')
    # Return null if key doesn't exist
    except KeyError:
        return None

# Setter method
def set(key, value):
  db[key] = value

# Delete method
def delete(key):
  try:
    del db[key]
  except KeyError:
    # Do nothing if key doesn't exist
    return



import pushy.lib.api as api
import pushy.lib.mqtt as mqtt
from pushy.util.config import config
import pushy.util.localStorage as localStorage

def listen():
    # Connect to MQTT endpoint
    mqtt.connect()

def isRegistered():
    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])
    
    # Both values must exist
    return token and tokenAuth

def validateDeviceCredentials():
    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])

    # Not registered yet?
    if not token or not tokenAuth:
        raise Exception('The device is not registered to receive push notifications.')

    # JSON post data
    postData = { 
        'token': token, 
        'auth': tokenAuth,
        'sdk': int(config['sdk']['version'])
    }

    try:
        # Authenticate the device
        response = api.post('/devices/auth', postData)
    except Exception as e:
        # Request failed
        raise Exception('The API request failed: ' + str(e))

    # Validate response
    if not response['success']:
        raise Exception('An unexpected response was received from the Pushy API.')

def register(options):
    # Make sure options is an object
    if not options or type(options) is not dict:
        options = {}

    # No App ID passed in via options?
    if not options['appId']:
        raise Exception('Please provide your Pushy App ID as per the documentation.')

    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])
    tokenAppId = localStorage.get(config['storageKeys']['tokenAppId'])

    # Check for new app ID different than the one we registered with
    if token and tokenAppId and options['appId'] and type(tokenAppId) is str and tokenAppId != options['appId']:
        token = None

    # Already registered?
    if token and tokenAuth:
        try:
            # Validate device credentials
            validateDeviceCredentials()

            # Attempt to connect
            mqtt.connect()

            # Device credentials are valid
            return token
        except Exception as e:
            # Ignore error and register new device
            print('Device validation error: ' + str(e))

    # JSON post data
    postData = {
        'appId': options['appId'],
        'sdk': int(config['sdk']['version']),
        'platform': config['sdk']['platform']
    }

    try:
        # Register the device
        response = api.post('/register', postData)
    except Exception as e:
        # Registration failed
        raise Exception('The API request failed: ' + str(e))

    # Validate response
    if not response['token'] or not response['auth']:
        raise Exception('An unexpected response was received from the Pushy API.')

    # Save device token and auth in local storage
    localStorage.set(config['storageKeys']['token'], response['token'])
    localStorage.set(config['storageKeys']['tokenAuth'], response['auth'])
    localStorage.set(config['storageKeys']['tokenAppId'], options['appId'])

    # Disconnect any existing connection
    mqtt.disconnect()

    # Start listening for notifications
    listen()

    # Provide app with device token
    return response['token']

def setNotificationListener(listener):
    # Store reference to listener
    mqtt.setNotificationListener(listener)
    
def setConnectionListener(listener):
    # Store reference to listener
    mqtt.setConnectionListener(listener)

def setDisconnectionListener(listener):
    # Store reference to listener
    mqtt.setDisconnectionListener(listener)

def loop_forever():
    # Avoid program termination
    mqtt.loop_forever()

def subscribe(topics):
    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])

    # Not registered yet?
    if not token or not tokenAuth:
        raise Exception('This device is not registered to receive push notifications.')

    # Convert single topic to array
    if type(topics) is str:
        topics = [topics]

    # JSON post data
    postData = { 
        'token': token, 
        'auth': tokenAuth, 
        'topics': topics 
    }

    try:
        # Subscribe to topic(s)
        response = api.post('/devices/subscribe', postData)
    except Exception as err:
        # Request failed
        raise Exception('The API request failed: ' + str(err))

    # Validate response
    if not response['success']:
        raise Exception('An unexpected response was received from the Pushy API.')

def unsubscribe(topics):
    # Attempt to fetch existing token and auth
    token = localStorage.get(config['storageKeys']['token'])
    tokenAuth = localStorage.get(config['storageKeys']['tokenAuth'])

    # Not registered yet?
    if not token or not tokenAuth:
        raise Exception('This device is not registered to receive push notifications.')

    # Convert single topic to array
    if type(topics) is str:
        topics = [topics]

    # JSON post data
    postData = { 
        'token': token, 
        'auth': tokenAuth, 
        'topics': topics 
    }

    try:
        # Unsubscribe from topic(s)
        response = api.post('/devices/unsubscribe', postData)
    except Exception as err:
        # Request failed
        raise Exception('The API request failed: ' + str(err))

    # Validate response
    if not response['success']:
        raise Exception('An unexpected response was received from the Pushy API.')

def setHeartbeatInterval(seconds):
    # Invalid keepalive value?
    if not seconds or not int(seconds) or seconds < 5:
        raise Exception('Please provide a valid heartbeat interval in seconds.')

    # Save keep alive interval
    localStorage.set(config['storageKeys']['keepAliveInterval'], str(seconds))

def isEnterpriseConfigured():
    # Check whether the key is set
    return localStorage.get(config['storageKeys']['enterpriseEndpoint']) != None

def setEnterpriseConfig(endpoint, mqttEndpoint):
    # Clear requested?
    if not endpoint or not mqttEndpoint:
        # Clear saved credentials and config
        localStorage.delete(config['storageKeys']['token'])
        localStorage.delete(config['storageKeys']['tokenAuth'])
        localStorage.delete(config['storageKeys']['tokenAppId'])
        localStorage.delete(config['storageKeys']['enterpriseEndpoint'])
        localStorage.delete(config['storageKeys']['enterpriseMqttEndpoint'])

        # Disconnect any existing connection and stop execution
        return mqtt.disconnect()

    # Strip trailing slash from API endpoint
    if endpoint[-1] == '/':
        endpoint = endpoint[0, endpoint.length - 1]

    # Strip trailing slash from mqtt endpoint
    if mqttEndpoint[-1] == '/':
        mqttEndpoint = mqttEndpoint[0, mqttEndpoint.length - 1]

    # Strip ssl:// protocol prefix from mqtt endpoint
    mqttEndpoint = mqttEndpoint.replace('ssl://', '')

    # Retrieve previous endpoint (may be null)
    previousEndpoint = localStorage.get(config['storageKeys']['enterpriseEndpoint'])
    previousMqttEndpoint = localStorage.get(config['storageKeys']['enterpriseMqttEndpoint'])

    # Endpoint changed?
    if endpoint != previousEndpoint or mqttEndpoint != previousMqttEndpoint:
        # Clear existing registration
        localStorage.delete(config['storageKeys']['token'])
        localStorage.delete(config['storageKeys']['tokenAuth'])
        localStorage.delete(config['storageKeys']['tokenAppId'])

        # Save updated Pushy Enterprise hostnames in local storage
        localStorage.set(config['storageKeys']['enterpriseEndpoint'], endpoint)
        localStorage.set(config['storageKeys']['enterpriseMqttEndpoint'], mqttEndpoint)

def disconnect():
    # Attempt to disconnect from Pushy
    mqtt.disconnect()


import pathlib
from setuptools import setup

setup(
    name='pushy-python',
    version='1.0.8',
    description='The official Pushy SDK for Python apps.',
    long_description=(pathlib.Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    url='https://github.com/pushy-me/pushy-python',
    author='Pushy',
    author_email='support@pushy.me',
    license='Apache 2.0',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    packages=['pushy', 'pushy.lib', 'pushy.util'],
    install_requires=['paho-mqtt']
)